# sRGB TRC en / decoder Ver. 1.5b
# (c)2023 - 24 Ohnishi, Yasuo 
# 2024.6
from ij import IJ, ImagePlus
from ij import plugin

from ij import WindowManager as WM
from ij.io import OpenDialog
from ij.plugin import ChannelSplitter, RGBStackMerge
from ij.plugin import Duplicator
# ij.plugin.RGBStackMerge
from ij.process import ImageConverter, ImageProcessor, FloatProcessor
from ij.gui import GenericDialog
from ij.plugin.filter import ThresholdToSelection
from loci.plugins import BF
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#-----------------
def set_Lmax(imp):
    # set Maximum value of Lightness
    BitDepth = imp.getBitDepth()
    if BitDepth == 32:
        L_max = imp.getDisplayRangeMax()
        if int(L_max) > 65536:
            L_max = 16777215
        elif int(L_max) > 256:
            L_max = 65535
        elif int(L_max) > 1.01:
            L_max = 255
        else:
            L_max = 1.0
    elif BitDepth == 16:
        L_max = 65535
    else: # 8 or 24 bit
        L_max = 255
    return L_max


#--- image type detection
def imageTypeDetection(imp):
    StackSize = imp.getStackSize()
    BitDepth = imp.getBitDepth()
    return StackSize, BitDepth


#-----------------
def imageBinarySmpl(impOrg, lowerV, upperV, roiColor, maskname):
    # fill roiColor between lowerV - upperV
    # make binary mask image with setThreshold
    # lowerV: lowerThreshold / upperV: upperThreshold (double)
    # roiColor: Color Value, ROI filled with (double)
    # maskname: name of mask image (string)
    imp = Duplicator().run(impOrg)
    ip = imp.getProcessor()
    mask = ip.createProcessor(ip.width, ip.height) # Background Value = 0.0
    mask.setValue(roiColor) 
                    # with which color ROI filled
    ip.setThreshold(lowerV, \
        upperV, ImageProcessor.NO_LUT_UPDATE)
    roi = ThresholdToSelection.run(imp)
    mask.setRoi(roi)
    mask.fill(roi)
    maskimp = ImagePlus(maskname, mask)
    imp.close()
    stat = maskimp.getStatistics()
    return maskimp

#----load image-------
def load_image():
    imgCurrent = WM.getCurrentImage()
    if not imgCurrent:
        od = OpenDialog("Choose a file")
        folder = od.getDirectory()
        filename = od.getFileName()
    # Always Open file with Bio-format Plugin
        if filename:
            imps = BF.openImagePlus(folder + filename) 
            imp0 = imps[0] #imp0: Original Image 
            imp0.show()
        else:
            IJ.showMessage("File is not selected.")
    else: # target image is already open
        imp0 = imgCurrent
        filename = imp0.getTitle()
        try:
            folder = imp0.getOriginalFileInfo().directory
        except:
            folder = IJ.getDirectory("Output_directory")
    return imp0, filename, folder
#----
#-----------------------------
def get_Choice():
                # dialog of selecting target channel for mask valid lightness range
    gd = GenericDialog("sRGB TRC en/decoder")
    gd.addMessage("encode: Linear to sRGB TRC")
    gd.addMessage("decode: sRGB TRC to Linear")
    processChoice = ["encode", "decode"]
    gd.addChoice("Select applying process", processChoice, processChoice[1])
  
    #---- show the dialog ----
    gd.showDialog()

    process =  gd.getNextChoice()
    if gd.wasCanceled():
        process = "cancel"

    return process

#--------------------------
def TRCencode(impsCh1, impsCh2, L_max, BitDepth, chName, n):
    impsCh3 = []
    shadowThld = 0.0031308
    cshadowThld = shadowThld * L_max
    for i in range(n):
        # non shadow region
        stat0 = impsCh1[i].getStatistics()
        impmask = imageBinarySmpl(impsCh1[i], cshadowThld, \
                                  L_max, L_max, "non_shadow_mask")
        ImageConverter(impsCh1[i]).convertToGray32()
        impsCh1[i].getProcessor().multiply(1.0 / L_max) # convert to 0.0 - 1.0
        impsCh1[i].setDisplayRange(0.0, 1.0)
        impsCh1[i].getProcessor().gamma(1.0 / 2.4)
        impsCh1[i].getProcessor().multiply(1.055)
        impsCh1[i].getProcessor().subtract(0.055)
        impsCh1[i].getProcessor().multiply(L_max) # convert to 0.0 - L_max
        impsCh1[i].setDisplayRange(0.0, L_max)
        if BitDepth == 8:
            ImageConverter(impsCh1[i]).convertToGray8()
        elif BitDepth == 16:
            ImageConverter(impsCh1[i]).convertToGray16()
        #================================================
        ic = plugin.ImageCalculator()
        imp_nonShadow =  ic.run("Min create", impmask, impsCh1[i])
            # imp_nonShadow: 0 to Threshold: black / Threshold to L_max: grayscale

        # shadow region
        if stat0.min <= cshadowThld:
#
            impmaskShadow = imageBinarySmpl(impsCh2[i], 0.0, cshadowThld, L_max, \      
                                        "shadow_mask")
                # mask: 0 to Threshold: white / Threshold to L_max: black
            impsCh2[i].setDisplayRange(0.0, L_max)
            impsCh2[i].getProcessor().multiply(12.92)
            imp_Shadow =  ic.run("Min create", impmaskShadow, impsCh2[i])
            # imp_Shadow: 0 to Threshold: grayscale / Threshold to L_max: black
        else:
            imp_Shadow = impsCh2[i].duplicate()
            imp_Shadow.getProcessor().setValue(0.0)
            imp_Shadow.getProcessor().fill()
#

        # shadow + non shadow    
        impTmp =  ic.run("Max create", imp_Shadow, imp_nonShadow)
        impsCh3.append(impTmp)
        impsCh3[i].setTitle(chName[i])
        impsCh3[i].show()
    impmask.close()
    impmaskShadow.close()
    imp_nonShadow.close()
    imp_Shadow.close()
    return impsCh3

#--------------------------
def TRCdecode(impsCh1, impsCh2, L_max, chName, n):
    impsCh3 = []
    shadowThld = 0.04045
    cshadowThld = shadowThld * L_max
    for i in range(n):
        stat0 = impsCh1[i].getStatistics()
        # non shadow region
        #    Threshold to L_max: white
        impmask = imageBinarySmpl(impsCh1[i], cshadowThld, \
                                  L_max, L_max, "shadow_mask")
            # mask: 0 to Threshold: black / Threshold to L_max: white

            # ROI: white back: black 
        impsCh1[i].setDisplayRange(0.0, L_max)
        ImageConverter(impsCh1[i]).convertToGray32()
        impsCh1[i].getProcessor().add(0.055 * L_max)

        impsCh1[i].getProcessor().multiply(1 / 1.055)
        impsCh1[i].getProcessor().multiply(1.0 / L_max)
        impsCh1[i].setDisplayRange(0.0, 1.0)
        impsCh1[i].getProcessor().gamma(2.4)
        impsCh1[i].getProcessor().multiply(L_max)
        impsCh1[i].setDisplayRange(0.0, L_max)
        if BitDepth == 8:
            ImageConverter(impsCh1[i]).convertToGray8()
        elif BitDepth == 16:
            ImageConverter(impsCh1[i]).convertToGray16()

        ic = plugin.ImageCalculator()
        imp_nonShadow =  ic.run("Min create", impmask, impsCh1[i])
            # imp_nonShadow: 0 to Threshold: black / Threshold to L_max: grayscale
        #IJ.saveAsTiff(imp_nonShadow, folder + filename + \
        #    "_sRGB_TRC_imp_nonShadow_test_" + str(i) +".tif")

        # shadow region    
        #    Threshold to L_max: white

        if stat0.min <= cshadowThld:
            impmaskShadow = imageBinarySmpl(impsCh2[i], 0.0, cshadowThld, L_max, \
                                        "non_shadow_mask")
                # mask: 0 to Threshold: white / Threshold to L_max: black
            impsCh2[i].setDisplayRange(0.0, L_max)
            impsCh2[i].getProcessor().multiply(1 / 12.92)
            imp_Shadow =  ic.run("Min create", impmaskShadow, impsCh2[i])
                # imp_Shadow: 0 to Threshold: grayscale / Threshold to L_max: black
        else:
            imp_Shadow = impsCh2[i].duplicate()
            imp_Shadow.getProcessor().setValue(0.0)
            imp_Shadow.getProcessor().fill()

        # shadow + non shadow    
        impTmp =  ic.run("Max create", imp_Shadow, imp_nonShadow)
        impsCh3.append(impTmp)
        impsCh3[i].setTitle(chName[i])
        impsCh3[i].show()

    impmask.close()
    impmaskShadow.close()
    imp_nonShadow.close()
    imp_Shadow.close()
    return impsCh3

#---  code for main ---
imp0, filename, folder = load_image()
StackSize, BitDepth = imageTypeDetection(imp0)
if BitDepth == 24: # color processor
    BitDepth = 8
process = get_Choice()
if process != "cancel":
    if filename:
        IJ.log("folder: " + folder)
        IJ.log("filename: " + filename)
        IJ.log("StackSize: " + str(StackSize))
        IJ.log("BitDepth: " + str(BitDepth))
    n = imp0.getStackSize()
    L_max = set_Lmax(imp0)
    imp1 = Duplicator().run(imp0)
    imp2 = Duplicator().run(imp0)
    impsCh1 = ChannelSplitter.split(imp1)
    impsCh2 = ChannelSplitter.split(imp2)

    for i in range(n):
        impsCh1[i].getProcessor().setLut(None)
        impsCh2[i].getProcessor().setLut(None)
    chName = ['R', 'G', 'B'] # Channel name
    if process == "encode":
        impsCh3 = TRCencode(impsCh1, impsCh2, L_max, BitDepth, chName, n)
    
    elif process == "decode":
        impsCh3 = TRCdecode(impsCh1, impsCh2, L_max, chName, n)
    
    if n == 3:
        imp3 = RGBStackMerge.mergeChannels(impsCh3, True)
        #imp3 = IJ.run("Merge Channels...", "c1=R c2=G c3=B create keep")
    else:
        imp3 = impsCh3[0]
    IJ.saveAsTiff(imp3, folder + filename +"_sRGB_TRC_" + process +"d.tif")
    imp3.show()    
    imp0.close()
    imp1.close()
    imp2.close()
    for i in range(n):
        impsCh1[i].close()
        impsCh2[i].close()
        impsCh3[i].close()
    IJ.showMessage("TRC Conversion Finished")
else:
    IJ.showMessage("Job canceled")
    imp0.close()

    
#---

